@@ -6,7 +6,16 @@ use 5.010;
 
 no if $] >= 5.018, warnings => 'experimental::smartmatch';
 
-our $VERSION = '1.00';
+our $VERSION = '1.01';
+
+# create CONSTANTS for different Return Types
+use constant {
+	TYPE_STOP       => 0,
+	TYPE_PREDICTION => 1,
+	TYPE_MESSAGE    => 2,
+	TYPE_BASE       => 3,
+	TYPE_URA        => 4,
+};
 
 use Carp qw(confess cluck);
 use DateTime;
@@ -39,13 +48,23 @@ sub new {
 		hide_past      => $opt{hide_past} // 1,
 		stop           => $opt{stop},
 		via            => $opt{via},
+		via_id         => $opt{via_id},
+		stop_id        => $opt{stop_id},
+		line_id        => $opt{line_id},
+		circle         => $opt{circle},
 		post           => {
-			ReturnList =>
-			  'lineid,linename,directionid,destinationtext,vehicleid,'
-			  . 'tripid,estimatedtime,stopid,stoppointname'
+
+			# for easier debugging ordered in the returned order
+			ReturnList => 'stoppointname,stopid,latitude,longitude,lineid,'
+			  . 'linename,directionid,destinationtext,vehicleid,tripid,'
+			  . 'estimatedtime'
 		},
 	};
 
+	if ( $opt{with_stops} ) {
+		$self->{post}{StopAlso} = 'True';
+	}
+
 	$self->{ura_instant_url}
 	  = $self->{ura_base} . '/instant_V' . $self->{ura_version};
 
@@ -54,6 +73,27 @@ sub new {
 	$ua->env_proxy;
 
 	if ( substr( $self->{ura_instant_url}, 0, 5 ) ne 'file:' ) {
+
+		# filter by stop_id only if full_routes is not set
+		if ( not $self->{full_routes} and $self->{stop_id} ) {
+			$self->{post}{StopID} = $self->{stop_id};
+
+			# filter for via as well to make via work
+			if ( defined $self->{via_id} ) {
+				$self->{post}{StopID} .= q{,} . $self->{via_id};
+			}
+		}
+
+		# filter by line
+		if ( $self->{line_id} ) {
+			$self->{post}{LineID} = $self->{line_id};
+		}
+
+		# filter for Stops in circle (lon,lat,dist)
+		if ( $self->{circle} ) {
+			$self->{post}{Circle} = $self->{circle};
+		}
+
 		$response = $ua->post( $self->{ura_instant_url}, $self->{post} );
 	}
 	else {
@@ -68,7 +108,7 @@ sub new {
 	my $raw_str = $response->decoded_content;
 
 	if ( $self->{developer_mode} ) {
-		say $raw_str;
+		say decode( 'UTF-8', $raw_str );
 	}
 
 	# Fix encoding in case we're running through test files
@@ -88,14 +128,35 @@ sub parse_raw_data {
 		$dep =~ s{^\[}{};
 		$dep =~ s{\]$}{};
 
-		# first field == 4 => version information, no departure
-		if ( substr( $dep, 0, 1 ) != 4 ) {
-			$csv->parse($dep);
-			my @fields = $csv->fields;
-			push( @{ $self->{raw_list} }, \@fields );
-			for my $i ( 1, 6 ) {
-				$fields[$i] = encode( 'UTF-8', $fields[$i] );
+		$csv->parse($dep);
+		my @fields = $csv->fields;
+
+		# encode all fields
+		for my $i ( 1, 11 ) {
+			$fields[$i] = encode( 'UTF-8', $fields[$i] );
+		}
+
+		push( @{ $self->{raw_list} }, \@fields );
+
+		my $type = $fields[0];
+
+		if ( $type == TYPE_STOP ) {
+			my $stop_name = $fields[1];
+			my $stop_id   = $fields[2];
+			my $longitude = $fields[3];
+			my $latitude  = $fields[4];
+
+			# create Stop Dict
+			if ( not exists $self->{stops}{$stop_id} ) {
+				$self->{stops}{$stop_id} = Travel::Status::DE::URA::Stop->new(
+					name      => decode( 'UTF-8', $stop_name ),
+					id        => $stop_id,
+					longitude => $longitude,
+					latitude  => $latitude,
+				);
 			}
+		}
+		if ( $type == TYPE_PREDICTION ) {
 			push( @{ $self->{stop_names} }, $fields[1] );
 		}
 	}
@@ -118,6 +179,12 @@ sub get_stop_by_name {
 	return ( grep { $_ =~ m{$name}i } @{ $self->{stop_names} } );
 }
 
+sub get_stops {
+	my ($self) = @_;
+
+	return $self->{stops};
+}
+
 sub errstr {
 	my ($self) = @_;
 
@@ -130,28 +197,45 @@ sub results {
 
 	my $full_routes = $opt{calculate_routes} // $self->{full_routes} // 0;
 	my $hide_past   = $opt{hide_past}        // $self->{hide_past}   // 1;
+	my $line_id     = $opt{line_id}          // $self->{line_id};
 	my $stop        = $opt{stop}             // $self->{stop};
+	my $stop_id     = $opt{stop_id}          // $self->{stop_id};
 	my $via         = $opt{via}              // $self->{via};
+	my $via_id      = $opt{via_id}           // $self->{via_id};
 
 	my $dt_now = $self->{datetime};
 	my $ts_now = $dt_now->epoch;
 
-	if ($via) {
+	if ( $via or $via_id ) {
 		$full_routes = 1;
 	}
 
 	for my $dep ( @{ $self->{raw_list} } ) {
 
 		my (
-			$u1, $stopname, $stopid,    $lineid, $linename,
-			$u2, $dest,     $vehicleid, $tripid, $timestamp
+			$type,     $stopname,  $stopid,   $longitude,
+			$latitude, $lineid,    $linename, $directionid,
+			$dest,     $vehicleid, $tripid,   $timestamp
 		) = @{$dep};
 		my ( @route_pre, @route_post );
 
+		# only work on Prediction informations
+		if ( $type != TYPE_PREDICTION ) {
+			next;
+		}
+
+		if ( $line_id and not( $lineid eq $line_id ) ) {
+			next;
+		}
+
 		if ( $stop and not( $stopname eq $stop ) ) {
 			next;
 		}
 
+		if ( $stop_id and not( $stopid eq $stop_id ) ) {
+			next;
+		}
+
 		if ( not $timestamp ) {
 			cluck("departure element without timestamp: $dep");
 			next;
@@ -170,8 +254,10 @@ sub results {
 		my $ts_dep = $dt_dep->epoch;
 
 		if ($full_routes) {
-			my @route = map { [ $_->[9] / 1000, $_->[1] ] }
-			  grep { $_->[8] == $tripid } @{ $self->{raw_list} };
+			my @route
+			  = map { [ $_->[11] / 1000, $_->[1], $_->[2], $_->[3], $_->[4] ] }
+			  grep { $_->[10] == $tripid }
+			  grep { $_->[0] == 1 } @{ $self->{raw_list} };
 
 			@route_pre  = grep { $_->[0] < $ts_dep } @route;
 			@route_post = grep { $_->[0] > $ts_dep } @route;
@@ -182,6 +268,12 @@ sub results {
 				next;
 			}
 
+			if ( $via_id
+				and none { $_->[2] eq $via_id } @route_post )
+			{
+				next;
+			}
+
 			if ($hide_past) {
 				@route_pre = grep { $_->[0] >= $ts_now } @route_pre;
 			}
@@ -199,7 +291,10 @@ sub results {
 						epoch     => $_->[0],
 						time_zone => 'Europe/Berlin'
 					),
-					name => decode( 'UTF-8', $_->[1] )
+					name      => decode( 'UTF-8', $_->[1] ),
+					id        => $_->[2],
+					longitude => $_->[3],
+					latitude  => $_->[4],
 				  )
 			} @route_pre;
 			@route_post = map {
@@ -208,7 +303,10 @@ sub results {
 						epoch     => $_->[0],
 						time_zone => 'Europe/Berlin'
 					),
-					name => decode( 'UTF-8', $_->[1] )
+					name      => decode( 'UTF-8', $_->[1] ),
+					id        => $_->[2],
+					longitude => $_->[3],
+					latitude  => $_->[4],
 				  )
 			} @route_post;
 		}
@@ -220,7 +318,7 @@ sub results {
 				dt_now      => $dt_now,
 				line        => $linename,
 				line_id     => $lineid,
-				destination => decode( 'UTF-8', $dest ),
+				destination => $dest,
 				route_pre   => [@route_pre],
 				route_post  => [@route_post],
 				stop        => $stopname,
@@ -264,7 +362,7 @@ API" data providers (e.g. ASEAG)
 
 =head1 VERSION
 
-version 1.00
+version 1.01
 
 =head1 DESCRIPTION
 
@@ -300,7 +398,7 @@ The request URL is I<ura_base>/instant_VI<version>, so for
 C<< http://ivu.aseag.de/interfaces/ura >>, C<< 1 >> this module will send
 requests to C<< http://ivu.aseag.de/interfaces/ura/instant_V1 >>.
 
-The following parameter is optional:
+All remaining parameters are optional.
 
 =over
 
@@ -309,6 +407,18 @@ The following parameter is optional:
 Passed on to C<< LWP::UserAgent->new >>. Defaults to C<< { timeout => 10 } >>,
 you can use an empty hashref to override it.
 
+=item B<circle> => I<lon,lat,dist>
+
+Only request departures for stops which are located up to I<dist> meters
+away from the location specified by I<lon> and I<lat>. Example parameter:
+"50.78496,6.10897,100".
+
+=item B<with_stops> => B<0>|B<1>
+
+When set to B<1> (or any other true value): Also request all stops satisfying
+the specified parameters. They can be accessed with B<get_stops>. Defaults to
+B<0>.
+
 =back
 
 Additionally, all options supported by C<< $status->results >> may be specified
@@ -328,6 +438,15 @@ Returns a list of stops matching I<$stopname>. For instance, if the stops
 parameter "bushof" will return "Aachen Bushof" and "Eupen Bushof", while
 "brand" will only return "Brand".
 
+=item $status->get_stops
+
+Returns a hash reference describing all distinct stops returned by the request.
+Each key is the unique ID of a stop and contains a
+Travel::Status::DE::URA::Stop(3pm) object describing it.
+
+Only works when $status was created with B<with_stops> set to a true value.
+Otherwise, undef is returned.
+
 =item $status->results(I<%opt>)
 
 Returns a list of Travel::Status::DE::URA::Result(3pm) objects, each describing
@@ -347,15 +466,28 @@ their B<route_> accessors. Otherwise, those will just return nothing
 
 Do not include past departures in the result list and the computed timetables.
 
+=item B<line_id> => I<ID>
+
+Only return departures of line I<ID>.
+
 =item B<stop> => I<name>
 
 Only return departures at stop I<name>.
 
+=item B<stop_id> => I<ID>
+
+Only return departures at stop I<ID>.
+
 =item B<via> => I<vianame>
 
 Only return departures containing I<vianame> in their route after their
 corresponding stop. Implies B<calculate_routes>=1.
 
+=item B<via_id> => I<ID>
+
+Only return departures containing I<ID> in their route after their
+corresponding stop. Implies B<calculate_routes>=1.
+
 =back
 
 =back
